from pyramid.session import (
    BaseCookieSessionFactory,
    PickleSerializer,
)

from .serializer import EncryptedSerializer

def EncryptedCookieSessionFactory(
    secret,
    cookie_name='session',
    max_age=None,
    path='/',
    domain=None,
    secure=False,
    httponly=False,
    timeout=1200,
    reissue_time=0,
    set_on_exception=True,
    serializer=None,
):
    """
    Configure a :term:`session factory` which will provide encrypted
    cookie-based sessions.  The return value of this
    function is a :term:`session factory` which may be used with the
    :meth:`pyramid.config.Configurator.set_session_factory` method.

    The session factory returned by this function will create sessions
    which are limited to storing fewer than 4000 bytes of data (as the
    payload must fit into a single cookie).

    Parameters:

    ``secret``
      A string which is used to sign the cookie. The secret should be at
      least as long as the block size of the selected hash algorithm. For
      ``sha512`` this would mean a 128 bit (64 character) secret.  It should
      be unique within the set of secret values provided to Pyramid for
      its various subsystems (see :ref:`admonishment_against_secret_sharing`).

    ``cookie_name``
      The name of the cookie used for sessioning. Default: ``'session'``.

    ``max_age``
      The maximum age of the cookie used for sessioning (in seconds).
      Default: ``None`` (browser scope).

    ``path``
      The path used for the session cookie. Default: ``'/'``.

    ``domain``
      The domain used for the session cookie.  Default: ``None`` (no domain).

    ``secure``
      The 'secure' flag of the session cookie. Default: ``False``.

    ``httponly``
      Hide the cookie from Javascript by setting the 'HttpOnly' flag of the
      session cookie. Default: ``False``.

    ``timeout``
      A number of seconds of inactivity before a session times out. If
      ``None`` then the cookie never expires. This lifetime only applies
      to the *value* within the cookie. Meaning that if the cookie expires
      due to a lower ``max_age``, then this setting has no effect.
      Default: ``1200``.

    ``reissue_time``
      The number of seconds that must pass before the cookie is automatically
      reissued as the result of accessing the session. The
      duration is measured as the number of seconds since the last session
      cookie was issued and 'now'.  If this value is ``0``, a new cookie
      will be reissued on every request accessing the session. If ``None``
      then the cookie's lifetime will never be extended.
      A good rule of thumb: if you want auto-expired cookies based on
      inactivity: set the ``timeout`` value to 1200 (20 mins) and set the
      ``reissue_time`` value to perhaps a tenth of the ``timeout`` value
      (120 or 2 mins).  It's nonsensical to set the ``timeout`` value lower
      than the ``reissue_time`` value, as the ticket will never be reissued.
      However, such a configuration is not explicitly prevented.
      Default: ``0``.

    ``set_on_exception``
      If ``True``, set a session cookie even if an exception occurs
      while rendering a view. Default: ``True``.

    ``serializer``
      An object with two methods: ``loads`` and ``dumps``.  The ``loads``
      method should accept bytes and return a Python object.  The ``dumps``
      method should accept a Python object and return bytes.  A ``ValueError``
      should be raised for malformed inputs.  If a serializer is not passed,
      the :class:`pyramid.session.PickleSerializer` serializer will be used.
    """
    if serializer is None:
        serializer = PickleSerializer()

    encrypted_serializer = EncryptedSerializer(
        secret,
        serializer=serializer,
    )

    return BaseCookieSessionFactory(
        encrypted_serializer,
        cookie_name=cookie_name,
        max_age=max_age,
        path=path,
        domain=domain,
        secure=secure,
        httponly=httponly,
        timeout=timeout,
        reissue_time=reissue_time,
        set_on_exception=set_on_exception,
    )



import binascii
import getopt
import sys

from nacl.secret import SecretBox
from nacl.utils import random


def generate_secret(as_hex=True):
    """Generate a random, 32-byte secret.

    :type as_hex: boolean
    :param as_hex: If true, convert the secret to hex.

    :rtype: bytes
    :returns: the secret (perhaps converted to hex).
    """
    secret = random(SecretBox.KEY_SIZE)
    if as_hex:
        secret = binascii.hexlify(secret)
    return secret


def print_secret():  # pragma: NO COVER
    """Print a generated, random secret, to standard output.

    The secret is a 32-byte random string, encoded using
    :func:`binascii.hexlify`.
    """
    def _usage(message=None, rc=1):
        print(print_secret.__doc__)
        if message is not None:
            print(message)
        sys.exit(rc)

    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], '?h', ['help'])
    except getopt.GetoptError as e:
        _usage(str(e))

    if args:
        _usage('No arguments allowed.')

    for k, v in opts:
        if k in ('-?', '-h', '--help'):
            _usage(rc=2)

    print(generate_secret())



from pyramid import testing
import unittest
import webtest

class TestEncryptedCookieSessionFactory(unittest.TestCase):
    default_secret = b'seekrit!' * 4  # 32 bytes

    def _makeOne(self, **kw):
        from ..session import EncryptedCookieSessionFactory
        kw.setdefault('secret', self.default_secret)
        return EncryptedCookieSessionFactory(**kw)

    def test_it(self):
        config = testing.setUp()
        factory = self._makeOne()
        config.set_session_factory(factory)
        state_found = []
        def begin_view(context, request):
            state_found.append(request.session.copy())
            request.session['state'] = 1
            return 'begin'
        def end_view(context, request):
            state_found.append(request.session.copy())
            return 'end'
        config.add_view(begin_view, name='begin', renderer='string')
        config.add_view(end_view, name='end', renderer='string')
        app = config.make_wsgi_app()
        testapp = webtest.TestApp(app)

        testapp.get('/begin')
        testapp.get('/end')
        self.assertTrue('session' in testapp.cookies)
        self.assertEqual(state_found[0], {})
        self.assertEqual(state_found[1], {'state': 1})



import unittest


class Test_generate_secret(unittest.TestCase):

    RANDOM = b'\x01\x12\x23\x34\x45'

    def _callFUT(self, *args, **kw):
        from ..scripts import generate_secret
        return generate_secret(*args, **kw)

    def test_implicit(self):
        from binascii import hexlify
        from .. import scripts as MUT

        class _SecretBox(object):
            KEY_SIZE = 32

        def _random(size):
            return self.RANDOM

        with _Monkey(MUT, SecretBox=_SecretBox, random=_random):
            secret = self._callFUT()

        self.assertEqual(secret, hexlify(self.RANDOM))

    def test_explicit(self):
        from .. import scripts as MUT

        class _SecretBox(object):
            KEY_SIZE = 32

        def _random(size):
            return self.RANDOM

        with _Monkey(MUT, SecretBox=_SecretBox, random=_random):
            secret = self._callFUT(as_hex=False)

        self.assertEqual(secret, self.RANDOM)



class _Monkey(object):
    # context-manager for replacing module names in the scope of a test.

    def __init__(self, module, **kw):
        self.module = module
        self.to_restore = dict([(key, getattr(module, key)) for key in kw])
        for key, value in kw.items():
            setattr(module, key, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for key, value in self.to_restore.items():
            setattr(self.module, key, value)



import unittest


class EncryptedSerializerTests(unittest.TestCase):

    def _getTargetClass(self):
        from ..serializer import EncryptedSerializer
        return EncryptedSerializer

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def _serialize(self, value, secret, nonce):
        import base64
        from pyramid.compat import pickle
        from nacl.secret import SecretBox

        cstruct = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
        fstruct = SecretBox(secret).encrypt(cstruct, nonce)
        return base64.urlsafe_b64encode(fstruct).rstrip(b'=')

    def test_ctor_w_invalid_key(self):
        SECRET = b'SEEKRIT!'
        self.assertRaises(ValueError, self._makeOne, SECRET)

    def test_dumps(self):
        from .. import serializer as MUT
        SECRET = b'SEEKRIT!' * 4  # 32 bytes
        NONCE = b'\x01' * 24
        APPSTRUCT = {'foo': 'bar'}
        expected = self._serialize(APPSTRUCT, SECRET, NONCE)

        eps = self._makeOne(SECRET)
        with _Monkey(MUT, random=lambda size: NONCE):
            encrypted = eps.dumps(APPSTRUCT)
        self.assertEqual(expected, encrypted)

    def test_loads(self):
        SECRET = b'SEEKRIT!' * 4  # 32 bytes
        NONCE = b'\x01' * 24
        APPSTRUCT = {'foo': 'bar'}
        CIPHERTEXT = self._serialize(APPSTRUCT, SECRET, NONCE)

        eps = self._makeOne(SECRET)
        loaded = eps.loads(CIPHERTEXT)
        self.assertEqual(loaded, APPSTRUCT)

    def test_loads_with_invalid_b64(self):
        eps = self._makeOne(b'SEEKRIT!' * 4)
        self.assertRaises(ValueError, eps.loads, b'not*b64*data')

    def test_loads_with_tampered_content(self):
        import base64
        from pyramid.compat import pickle
        from nacl.secret import SecretBox

        secret = b'SEEKRIT!' * 4
        appstruct = {'foo': 'bar'}
        nonce = b'\x01' * 24
        cstruct = pickle.dumps(appstruct, pickle.HIGHEST_PROTOCOL)
        fstruct = SecretBox(secret).encrypt(cstruct, nonce)
        fstruct = b'tamper' + fstruct
        bstruct = base64.urlsafe_b64encode(fstruct).rstrip(b'=')

        eps = self._makeOne(b'SEEKRIT!' * 4)
        self.assertRaises(ValueError, eps.loads, bstruct)


class _Monkey(object):
    # context-manager for replacing module names in the scope of a test.

    def __init__(self, module, **kw):
        self.module = module
        self.to_restore = dict([(key, getattr(module, key)) for key in kw])
        for key, value in kw.items():
            setattr(module, key, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for key, value in self.to_restore.items():
            setattr(self.module, key, value)



from base64 import urlsafe_b64decode
from base64 import urlsafe_b64encode
import binascii

from nacl.exceptions import CryptoError
from nacl.secret import SecretBox
from nacl.utils import random
from pyramid.session import PickleSerializer


class EncryptedSerializer(object):
    """Encrypt session state using PyNaCl.

    :type secret: bytes
    :param secret: a 32-byte random secret for encrypting/decrypting the
                   pickled session state.
    :param serializer:
        An object with two methods: ``loads`` and ``dumps``. The ``loads``
        method should accept bytes and return a Python object. The ``dumps``
        method should accept a Python object and return bytes. A ``ValueError``
        should be raised for malformed inputs. Default: ``None``, which will
        use :class:`pyramid.session.PickleSerializer`.
    """
    def __init__(self, secret, serializer=None):
        if len(secret) != SecretBox.KEY_SIZE:
            raise ValueError(
                "Secret should be a random bytes string of length %d" %
                SecretBox.KEY_SIZE)
        self.box = SecretBox(secret)

        if serializer is None:
            serializer = PickleSerializer()

        self.serializer = serializer

    def loads(self, bstruct):
        """Decrypt session state.

        :type encrypted_state: bytes
        :param encrypted_state: the encrypted session state.

        :rtype: :class:`dict` / picklable mapping
        :returns: the decrypted, unpickled session state, as passed as
                  ``session_state`` to :meth:`dumps`.
        """
        try:
            b64padding = b'=' * (-len(bstruct) % 4)
            fstruct = urlsafe_b64decode(bstruct + b64padding)
        except (binascii.Error, TypeError) as e:
            raise ValueError('Badly formed base64 data: %s' % e)

        try:
            payload = self.box.decrypt(fstruct)
        except CryptoError as e:
            raise ValueError('Possible tampering: %s' % e)
        return self.serializer.loads(payload)

    def dumps(self, session_state):
        """Encrypt session state.

        :type session_state: :class:`dict` / picklable mapping
        :param session_state: the session state to be encrypted.

        :rtype: bytes
        :returns: the encrypted session state
        """
        cstruct = self.serializer.dumps(session_state)
        nonce = random(SecretBox.NONCE_SIZE)
        fstruct = self.box.encrypt(cstruct, nonce)
        return urlsafe_b64encode(fstruct).rstrip(b'=')



from setuptools import find_packages
from setuptools import setup


with open('README.rst') as f:
    README = f.read()

with open('CHANGES.rst') as f:
    CHANGES = f.read()

tests_require = ['WebTest']

setup(name='pyramid_nacl_session',
      version='0.3',
      description='Encrypted sessison cookie serializer ofr Pyramid',
      long_description='\n\n'.join([README, CHANGES]),
      url='https://github.com/Pylons/pyramid_nacl_session',
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "License :: Repoze Public License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP :: Session",
        "Topic :: Security :: Cryptography",
      ],
      packages=find_packages(),
      include_package_data=True,
      install_requires=[
        'pyramid>=1.5',
        'PyNaCl',
      ],
      extras_require={
          'testing': tests_require + [
              'nose',
              'coverage',
          ],
          'docs': [
              'Sphinx',
              'pylons-sphinx-themes',
          ],
      },
      test_suite='pyramid_nacl_session.tests',
      tests_require=tests_require,
      entry_points = """\
      [console_scripts]
      print_secret = pyramid_nacl_session.scripts:print_secret
      """
)



def _export(thing):
    """ No-op; document that ``thing`` is an API.

    Also, shuts pyflakes up about unused import.
    """

from .serializer import EncryptedSerializer
_export(EncryptedSerializer)
from .session import EncryptedCookieSessionFactory
_export(EncryptedCookieSessionFactory)
from .scripts import generate_secret
_export(generate_secret)
